#!/usr/bin/env python

import sfml as sf
import random

from tools import *

try:
    import rotary
    USE_MOUSE = False
except SystemExit:
    USE_MOUSE = True

FPS = 45

################################################################################

class BoundedBall:
    ''' This class represents a ball bounded within a circle of radius 1.'''
    def __init__(self, loc = sf.Vector2f(0, 0),
                 vel = sf.Vector2f(0, 0), radius = 0.04):
        self.loc = loc
        self.vel = vel
        self.radius = radius
        self.increment = norm(self.vel) / 10
        self.bounces = 0
        
    def update(self, dT, check, bounce_callback,
               death_callback, sigma = 0.3):
        ''' Updates the ball's position and velocity for a timestep dT.
            In the event of a collision with the wall, the angle of incidence
            equals the angle of reflection plus a random value chosen from a
            Gaussian distribution with standard deviation sigma.
            
            check is a function of one argument that returns True if an angle
            is blocked, False otherwise.'''
        
        speed = norm(self.vel)
        # Calculate the edge nearest to the wall
        edge = sf.Vector2f(self.loc.x + self.vel.x / speed * self.radius,
                           self.loc.y + self.vel.y / speed * self.radius)
        
        # Calculate how long it will take to hit the wall
        dT_collision = -(self.vel.x*edge.x + self.vel.y*edge.y - 
            sqrt(-self.vel.x**2*edge.y**2 +
                  2*self.vel.x*self.vel.y*edge.x*edge.y - 
                  self.vel.y**2*edge.x**2 + self.vel.x**2 +
                  self.vel.y**2)) / (self.vel.x**2 + self.vel.y**2)
                  
        # If we won't collide in this timestep, then continue onward
        if dT_collision > dT:
            self.loc += self.vel * dT
            return True    
        
        # Update for the first fraction of the timestep (up to the wall)
        self.loc += self.vel * dT_collision
        
        # Calculate the new angle
        vel_angle = atan2(self.vel.y, self.vel.x)
        incidence_angle = atan2(self.loc.y, self.loc.x)

        # If the ball is not blocked, then it dies!
        if not check(atan2(self.loc.y, self.loc.x)):
             self.loc += self.vel * (dT - dT_collision)
             death_callback(self.loc)
             return False
        else:
            self.bounces += 1 
            bounce_loc = sf.Vector2f(self.loc.x + self.vel.x / speed * self.radius,
                                     self.loc.y + self.vel.y / speed * self.radius)
            bounce_callback(bounce_loc)
            
        wall_angle = incidence_angle + pi/2
        reflection_angle = wall_angle
        reflection_angle = 2*wall_angle - vel_angle + random.gauss(0, sigma)
        
        # Calculate the reflected velocity
        speed += self.increment
        self.vel = sf.Vector2f(speed*cos(reflection_angle),
                               speed*sin(reflection_angle))
        
        # Update for the second fraction of the timestep
        self.loc += self.vel * (dT - dT_collision)
        return True
        
    def draw(self, target, states):
        ''' Draws the ball as a white circle with radius given
            by self.radius.  If failed is true, the ball is red instead
            of white.'''
        circ = sf.CircleShape(self.radius * 0.9, 100)
        circ.move(-self.radius + self.loc.x, -self.radius + self.loc.y)
        circ.outline_thickness = self.radius * 0.1
        circ.outline_color = sf.Color(128,128,128)
            
        target.draw(circ)

################################################################################

class Paddle(sf.Shape):
    def __init__(self, angle = 0, width = pi/8, resolution = 10):
        sf.Shape.__init__(self)
        self.angle = angle
        self.width = width
        self.height = 0.2/5.0
        self.resolution = resolution
        
        self.outline_thickness = 0.03
        self.outline_color = sf.Color(255, 255, 255)
        self.fill_color = sf.Color(0, 0, 0, 0)

    def get_point_count(self):
        return self.resolution * 2
    
    def get_point(self, index):
        ''' Returns a point on the paddle.  Points 0 to resolution - 1 are
            moving around the circle clockwise, resolution - 2*resolution-1
            are moving back counterclockwise.'''
        if index < self.resolution:
            a = self.angle - self.width/2 + (self.width * index / float(self.resolution - 1))
        else:
            a = self.angle + self.width/2 - self.width* (index - self.resolution) / float(self.resolution - 1)
        r = 1 + self.outline_thickness
        return sf.Vector2f(r*cos(a), r*sin(a))
            
    def set_angle(self, angle):
        ''' Sets the paddle's angle, stored as a radian from -pi to pi.'''
        self.angle = angle
        self.update()
    
    def check(self, angle):
        ''' Returns true if the argument angle is covered by the paddle. '''
        lower = self.angle - self.width/2
        upper = self.angle + self.width/2 

        # Standard case
        if lower > -pi and upper < pi:
            return lower < angle < upper
        # Special cases for wrapping around
        elif lower < -pi:
            return lower + 2*pi < angle or upper > angle
        elif upper > pi:
            return lower < angle or upper - 2*pi > angle

################################################################################

class PolarRect(sf.Shape):
    ''' A rectangle in polar coordinates.'''
    
    def __init__(self, angle, radius, width, height):
        ''' angle is center angle.
            width is angle occupied by rectangle
            radius is radius of rectangle's center
            height is rectangle's radial height'''
        sf.Shape.__init__(self)
        
        self.angle = angle
        self.radius = radius
        self.width = width
        self.height = height
        self.update()
        
    def get_point_count(self):
        return 4
    
    def get_point(self, index):
        lower = self.angle - self.width/2
        upper = self.angle + self.width/2
        if index == 0:
            return sf.Vector2f(self.radius*cos(lower), self.radius*sin(lower))
        elif index == 1:
            return sf.Vector2f(self.radius*cos(upper), self.radius*sin(upper))
        elif index == 2:
            return sf.Vector2f((self.radius+self.height)*cos(upper),
                               (self.radius+self.height)*sin(upper))    
        elif index == 3:
            return sf.Vector2f((self.radius+self.height)*cos(lower),
                               (self.radius+self.height)*sin(lower))    

################################################################################

class Particle:
    ''' Simple container for a circular particle. '''
    def __init__(self, loc, vel, radius, lifetime,
                 color = sf.Color(255, 255, 255, 128)):
        self.loc = loc.copy()
        self.vel = vel.copy()
        self.radius = radius
        self.lifetime = lifetime
        self.life = lifetime
        self.color = color

    def draw(self, target, states):
        ''' Draws an individual particle.'''
        radius = self.radius * self.life / float(self.lifetime)
        circ = sf.CircleShape(radius)
        circ.fill_color = self.color
        circ.move(self.loc.x - radius/2, self.loc.y - radius/2)
        target.draw(circ)
    
    def update(self, dT):
        ''' Updates particle's location and life.  Returns False if the
            particle is dead, True if the particle is still alive.'''
        self.loc += self.vel * dT
        self.life -= 1
        if not self.life:
            return False
        return True
        
################################################################################
class PongGame:
    def __init__(self, fullscreen = False):
        ''' Create a new instance of the PongGame class.  This class
            maintains the game's state.'''
            
        if fullscreen:
            self.window = sf.RenderWindow(
                sf.VideoMode.get_fullscreen_modes()[0],
                'PongPong',
                sf.Style.FULLSCREEN,
                sf.ContextSettings(antialiasing = 32))
        else:
            self.window = sf.RenderWindow(
               sf.VideoMode(600, 375),
               'PongPong',
               sf.Style.CLOSE,
               sf.ContextSettings(antialiasing = 32))
        
        self.window.vertical_sync_enabled = True
        self.window.mouse_cursor_visible = False
        
        # Figure out what the game's view should be
        aspect_ratio = self.window.width / float(self.window.height)
        extra = 1.4
        self.pong_view = sf.View.from_rect(
            sf.FloatRect(-aspect_ratio * extra, extra, 
                         aspect_ratio * 2 * extra, -2*extra))
        self.window.view = self.pong_view
        
        # Create the actor objects
        self.ball = BoundedBall(loc = sf.Vector2f(0, -0.5), vel = sf.Vector2f(0, 1))
        self.paddle = Paddle()
        
        self.mouse_loc = (0,0)
        self.particles = []
        
        self.state = 'loading'
                
        # Draw once to disguise loading time
        self.draw()
        self.load_resources()
        self.state = 'starting'
        
        self.clock = sf.Clock()
        self.last_time = self.clock.elapsed_time
    
    def load_resources(self):
        '''Loads sound resources.'''
    
        # Load the sound effects
        die_sound_buff = sf.SoundBuffer.load_from_file('die.wav')
        bounce_sound_buff = sf.SoundBuffer.load_from_file('bounce.wav')
        self.die_sound = sf.Sound(die_sound_buff)
        self.bounce_sound = sf.Sound(bounce_sound_buff)
        
    def restart(self):
        ''' Resets the system to start playing again.'''
        self.ball.loc = sf.Vector2f(0, -0.5)
        self.ball.vel = sf.Vector2f(0, 1)
        self.ball.bounces = 0
        self.ball.dead = False
        self.state = 'playing'
        self.paddle.outline_color = sf.Color(255, 255, 255)
    
    def handle_events(self):
        ''' Iterate through window events.  Stops main loop if the window
            is closed, escape is pressed, or command+Q is pressed.'''
        # Check if the GUI window is still open
        for event in self.window.iter_events():
            if event.type == sf.Event.CLOSED:
                self.state = 'quit'
            elif event.type == sf.Event.KEY_PRESSED:
                if event.code == sf.Keyboard.ESCAPE:
                    self.state = 'quit'
                elif event.system and event.code == sf.Keyboard.Q:
                    self.state = 'quit'
                elif self.state == 'starting':
                    self.state = 'playing'
                elif self.state == 'lost':
                    self.restart()
                    self.state = 'playing'
            elif USE_MOUSE and event.type == sf.Event.MOUSE_MOVED:
                self.mouse_loc = self.window.convert_coords(event.x, event.y)
                self.paddle.set_angle(atan2(self.mouse_loc[1], self.mouse_loc[0]))
    
    def draw_ui(self):
        ''' Draws status messages, based on the game state.'''
        if self.state == 'playing':
            t = sf.Text("Score: %i" % self.ball.bounces,
                        character_size = self.window.height/20.)
            t.move(t.local_bounds.height/2.0, t.local_bounds.height/2.0)
            self.window.draw(t)

        if self.state == 'loading':
            t = sf.Text("Loading...",
                        character_size = self.window.height/20.)
            t.move(self.window.width/2 - t.local_bounds.width/2,
                   self.window.height/2 - t.local_bounds.height)
            self.window.draw(t)
        
        if self.state == 'starting':
            t = sf.Text("Press any key to begin.",
                        character_size = self.window.height/20.)
            t.move(self.window.width/2 - t.local_bounds.width/2,
                   self.window.height/2 - t.local_bounds.height)
            self.window.draw(t)
        
        if self.state == 'lost':
            t = sf.Text(
                "Total score: %i\n\nPress any key to replay." % self.ball.bounces,
                character_size = self.window.height/20.)
            t.move(self.window.width/2 - t.local_bounds.width/2,
                   self.window.height/2 - t.local_bounds.height/2)
            self.window.draw(t)
    
    def draw_field(self):
        ''' Draws the playing field, including a circle,
            outer tick marks at 10 degree intervals, and inner
            tick marks at 10 degree intervals.'''
        circ = sf.CircleShape(1, 100)
        circ.move(-1, -1)
        circ.outline_thickness = 0.01
        if self.state == 'lost':
            circ.outline_color = sf.Color(50, 50, 50)
        else:
            circ.outline_color = sf.Color(128, 128, 128)
        circ.fill_color = sf.Color(0, 0, 0, 0)
        self.window.draw(circ)
        
        # Draw outer marks
        for deg in range(0, 360, 10):
            rad = deg * pi/180 + pi/2
            if deg % 360 == 0:
                height = 0.2
                alpha = 255
            elif deg % 30 == 0:
                height = 0.12
                alpha = 128
            else:
                height = 0.08
                alpha = 64
            
            rect = PolarRect(rad, 1.05, 0.02, height)
            rect.fill_color = sf.Color(255, 255, 255, alpha)
            self.window.draw(rect)
            
        # Draw inner marks
        for deg in range(0, 360, 10):
            rad = deg * pi/180 + self.paddle.angle
            if deg % 360 == 0:
                height = 0.2
                alpha = 255
            elif deg % 30 == 0:
                height = 0.12
                alpha = 128
            else:
                height = 0.08
                alpha = 64
                
            rect = PolarRect(rad, 0.95 - height, 0.02, height)
            rect.fill_color = sf.Color(255, 255, 255, alpha)
            self.window.draw(rect)
            
    def draw(self):
        ''' Draws the game's state.'''
        self.window.clear(sf.Color.BLACK)
             
        self.draw_field() 
        
        if self.state == 'lost':
            self.paddle.outline_color = sf.Color(100, 100, 100)
        self.window.draw(self.paddle)
        
        if self.state != 'lost':
            self.window.draw(self.ball)

        for p in self.particles:
            self.window.draw(p)
            
        # Draw the mouse as a green diamond
        if USE_MOUSE:
            circ = sf.CircleShape(0.025, 4)
            circ.move(self.mouse_loc[0], self.mouse_loc[1])
            circ.move(0.025*sin(self.paddle.angle),
                     -0.025*cos(self.paddle.angle))
            circ.rotate(self.paddle.angle*180/pi)
            circ.fill_color = sf.Color(0, 255, 0)
            self.window.draw(circ)
            
        # Load the default view (top left is 0, 0; window width is size)
        self.window.view = self.window.default_view
        self.draw_ui()
        self.window.view = self.pong_view
        
        self.window.display()
    
    def idle(self):
        ''' Kills time until 1/FPS has passed, then returns.
        
            If the rotary sensor is plugged in, samples sensor and
            updates paddle angle.'''
        while (self.clock.elapsed_time - self.last_time).as_seconds() < 1/float(FPS):
            if not USE_MOUSE:            
                paddle_angle = rotary.get_angle()
                self.paddle.set_angle(paddle_angle)
        self.last_time += sf.Time(1/float(FPS))

    def update_particles(self, dT):
        self.particles = filter(lambda p: p.update(dT), self.particles)
        
    def bounce_callback(self, bounce_loc):
        ''' This function should be called when the ball bounces.
            It plays the bounce sound and spawns a set of particles
            at the bounce location.'''
        self.bounce_sound.play()
        for i in range(10):
            angle = random.uniform(0, 2*pi)
            particle = Particle(bounce_loc,
                                sf.Vector2f(0.2*cos(angle), 0.2*sin(angle)),
                                0.01, 10,
                                sf.Color(128, 128, 128))
            self.particles += [particle]
    
    def death_callback(self, death_loc):
        ''' This function should be called when the ball escapes.
            It plays the death sound, sets the state to 'lost', and
            spawns a set of particles at the death location.'''
        self.state = 'lost'
        self.die_sound.play()
        for i in range(36):
            angle = i*10*pi/180
            particle = Particle(death_loc,
                                sf.Vector2f(0.4*cos(angle), 0.4*sin(angle)),
                                0.04, 45, sf.Color(255, 0, 0, 100))
            self.particles += [particle]
        
    def loop(self):
    
        # System timestep, empirically tweaked based on ball speed
        dT = 0.60 / FPS
    
        while self.state != 'quit':
            self.handle_events()
            self.idle()
            if self.state == 'playing':
                self.ball.update(dT, self.paddle.check,
                                 self.bounce_callback,
                                 self.death_callback)
            self.update_particles(dT)
            self.draw()

if __name__ == '__main__':
    pg = PongGame(fullscreen = True)
    pg.loop()